from typing import Union

import numpy

from judo.judo_backend import Backend, torch


DATA_TYPE_NAMES = {
    "bool",
    "int",
    "uint8",
    "int16",
    "int32",
    "int64",
    "float",
    "float16",
    "float32",
    "float64",
    "hash_type",
}

_int = int
_bool = bool
_float = float


def bool():
    funcs = {"numpy": lambda x: numpy.bool_, "torch": lambda x: torch.bool}
    return Backend.execute(None, funcs)


def int():
    funcs = {"numpy": lambda x: numpy.int64, "torch": lambda x: torch.int64}
    return Backend.execute(None, funcs)


def uint8():
    funcs = {"numpy": lambda x: numpy.uint8, "torch": lambda x: torch.uint8}
    return Backend.execute(None, funcs)


def int16():
    funcs = {"numpy": lambda x: numpy.int16, "torch": lambda x: torch.int16}
    return Backend.execute(None, funcs)


def int32():
    funcs = {"numpy": lambda x: numpy.int32, "torch": lambda x: torch.int32}
    return Backend.execute(None, funcs)


def int64():
    funcs = {"numpy": lambda x: numpy.int64, "torch": lambda x: torch.int64}
    return Backend.execute(None, funcs)


def float():
    funcs = {"numpy": lambda x: numpy.float32, "torch": lambda x: torch.float32}
    return Backend.execute(None, funcs)


def float16():
    funcs = {"numpy": lambda x: numpy.float16, "torch": lambda x: torch.float16}
    return Backend.execute(None, funcs)


def float32():
    funcs = {"numpy": lambda x: numpy.float32, "torch": lambda x: torch.float32}
    return Backend.execute(None, funcs)


def float64():
    funcs = {"numpy": lambda x: numpy.float64, "torch": lambda x: torch.float64}
    return Backend.execute(None, funcs)


def hash_type():
    funcs = {
        "numpy": lambda x: numpy.dtype("<U64") if Backend.use_true_hash() else numpy.int64,
        "torch": lambda x: torch.int64,
    }
    return Backend.execute(None, funcs)


class MetaScalar(type):
    @property
    def bool(cls):
        return bool()

    @property
    def uint8(cls):
        return uint8()

    @property
    def int16(cls):
        return int16()

    @property
    def int32(cls):
        return int32()

    @property
    def int64(cls):
        return int64()

    @property
    def int(cls):
        return int()

    @property
    def float(cls):
        return float()

    @property
    def float16(cls):
        return float16()

    @property
    def float32(cls):
        return float32()

    @property
    def float64(cls):
        return float64()

    @property
    def hash_type(cls):
        return hash_type()


class dtype(metaclass=MetaScalar):
    @classmethod
    def is_bool(cls, x):
        dtypes = (_bool, dtype.bool)
        if cls.is_tensor(x):
            return x.dtype in dtypes
        return isinstance(x, dtypes)

    @classmethod
    def is_float(cls, x):
        dtypes = (_float, cls.float64, cls.float32, cls.float16)
        if cls.is_tensor(x):
            return x.dtype in dtypes
        return isinstance(x, dtypes)

    @classmethod
    def is_int(cls, x):
        dtypes = (_int, dtype.int64, dtype.int32, dtype.int16)
        if cls.is_tensor(x):
            return x.dtype in dtypes
        return isinstance(x, dtypes)

    @classmethod
    def is_tensor(cls, x):
        return isinstance(x, (numpy.ndarray, torch.Tensor))  # or cls.is_hash_tensor(x)


class MetaTyping(type):
    @property
    def int(self):
        try:
            return Union[_int, dtype.int64, dtype.int32, dtype.int16]
        except Exception:
            return int

    @property
    def float(self):
        try:
            return Union[_float, dtype.float64, dtype.float32, dtype.float16]
        except Exception:
            return float

    @property
    def bool(self):
        try:
            return Union[_bool, dtype.bool]
        except Exception:
            return bool

    @property
    def Scalar(self):
        try:
            return Union[self.float, self.int]
        except Exception:
            return Union[_int, float]



try:
    from torch.multiprocessing.pool import Pool
except ImportError:
    from multiprocessing.pool import Pool

import numpy
import xxhash

import judo
from judo.judo_backend import Backend


class Hasher:

    _true_hash = bool(Backend.use_true_hash())

    def __init__(self, seed: int = 0):
        self._seed = seed
        if self._true_hash:
            self.pool = Pool()

    @property
    def uses_true_hash(self) -> bool:
        return self._true_hash

    @staticmethod
    def hash_numpy(x: numpy.ndarray) -> int:
        """Return a value that uniquely identifies a numpy array."""
        x = x.astype("|S576") if x.dtype == "O" else x
        return xxhash.xxh64_hexdigest(x.tobytes())

    @staticmethod
    def hash_torch(x):
        bytes = judo.to_numpy(x).tobytes()
        return xxhash.xxh32_intdigest(bytes)

    @classmethod
    def true_hash_tensor(cls, x):
        funcs = {
            "numpy": cls.hash_numpy,
            "torch": cls.hash_torch,
        }
        return Backend.execute(x, funcs)

    def __del__(self):
        self.pool.close()

    def get_one_id(self):
        self._seed += 1
        return self._seed

    def get_array_of_ids(self, n: int):
        ids = numpy.arange(n) + self._seed + 1
        self._seed += n + 1
        return judo.as_tensor(ids)

    def hash_tensor(self, x):
        if self._true_hash:
            return self.true_hash_tensor(x)
        return 0

    def hash_walkers(self, x):
        if self._true_hash:
            # hashes = self.pool.map(self.true_hash_tensor, x)
            hashes = [self.true_hash_tensor(x_i) for x_i in x]
            return judo.as_tensor(hashes)
        return self.get_array_of_ids(x.shape[0])

    def hash_state(self, state):
        if self._true_hash:
            _hash = hash(
                tuple(
                    [
                        self.hash_tensor(x) if k in state._tensor_names else hash(x)
                        for k, x in state.items()
                    ],
                ),
            )
            return _hash
        return 0


hasher = Hasher()



"""Functions that define a common api for numpy and pytorch functionality."""
from judo.functions import batching, fractalai, images, notebook, numpy, pytorch
from judo.functions.hashing import hasher
from judo.functions.random import random_state



from typing import Dict, Generator, Tuple, Union

import numpy

import judo
from judo.typing import Tensor


AVAILABLE_FUNCTIONS = {
    "similiar_chunks_indexes",
    "split_similar_chunks",
    "split_kwargs_in_chunks",
    "split_args_in_chunks",
}


def similiar_chunks_indexes(
    n_values,
    n_chunks,
    allow_size_1: bool = True,
) -> Generator[Tuple[int, int], None, None]:
    """
    Return the indexes for splitting an array in similar chunks.

    Args:
        n_values: Length of the array that will be split.
        n_chunks: Number of similar chunks.
        allow_size_1: If True chunks of size 1 can be returned. If false return \
                     indexes that avoid batches of size 1.

    Returns:
        Generator containing the indexes of every new chunk.

    """
    chunk_size = int(numpy.ceil(n_values / n_chunks))
    for i in range(0, n_values, chunk_size):
        if not allow_size_1 and i + chunk_size >= n_values - 2:  # Avoid batches of size 1
            yield i, i + chunk_size + 1
            break
        else:
            yield i, i + chunk_size


def split_similar_chunks(
    vector: Union[list, Tensor],
    n_chunks: int,
    allow_size_1: bool = True,
) -> Generator[Union[list, Tensor], None, None]:
    """
    Split an indexable object into similar chunks.

    Args:
        vector: Target object to be split.
        n_chunks: Number of similar chunks.
        allow_size_1: If True chunks of size 1 can be returned. If false return \
                     indexes that avoid batches of size 1.

    Returns:
        Generator that returns the chunks created after splitting the target object.

    """
    for start, end in similiar_chunks_indexes(len(vector), n_chunks, allow_size_1=allow_size_1):
        yield vector[start:end]


def split_kwargs_in_chunks(
    kwargs: Dict[str, Union[list, Tensor]],
    n_chunks: int,
    allow_size_1: bool = True,
) -> Generator[Dict[str, Union[list, Tensor]], None, None]:
    """Split the kwargs passed to ``make_transitions`` in similar batches."""
    n_values = len(next(iter(kwargs.values())))  # Assumes all data have the same len
    chunk_size = int(numpy.ceil(n_values / n_chunks))
    for start, end in similiar_chunks_indexes(n_values, n_chunks, allow_size_1=allow_size_1):
        if start + chunk_size >= n_values - 2:  # Do not allow the last chunk to have size 1
            yield {k: v[start:n_values] if judo.is_tensor(v) else v for k, v in kwargs.items()}
            break
        else:
            yield {k: v[start:end] if judo.is_tensor(v) else v for k, v in kwargs.items()}


def split_args_in_chunks(
    args: Tuple[Union[list, Tensor]],
    n_chunks: int,
    allow_size_1: bool = True,
) -> Generator[Union[list, Tensor], None, None]:
    """Split the args passed to ``make_transitions`` in similar batches."""
    n_values = len(args[0])
    for start, end in similiar_chunks_indexes(n_values, n_chunks, allow_size_1=allow_size_1):
        yield tuple(v[start:end] for v in args)



from judo.data_types import dtype
from judo.judo_backend import torch
from judo.judo_tensor import astype, tensor


AVAILABLE_FUNCTIONS = [
    "argmax",
    "hash_numpy",
    "hash_tensor",
    "concatenate",
    "stack",
    "clip",
    "repeat",
    "min",
    "max",
    "norm",
    "unsqueeze",
    "where",
    "sqrt",
    "tile",
    "logical_or",
    "logical_and",
]


def parse_dim(axis, dim):
    if axis is not None:
        return axis
    return dim


def concatenate(x, axis=0, out=None, dim=0):
    axis = parse_dim(axis, dim)
    return torch.cat(x, dim=axis, out=out)


def stack(x, axis=0, out=None, dim=0):
    axis = parse_dim(axis, dim)
    return torch.stack(x, dim=axis, out=out)


def clip(x, a_min, a_max, out=None):

    _tensor = min(x, other=astype(tensor(a_max), dtype=x.dtype))
    return max(_tensor, other=astype(tensor(a_min), dtype=x.dtype), out=out)


def repeat(x, repeat, axis=None, dim=None):
    axis = parse_dim(axis, dim)
    return torch.repeat_interleave(x, repeat, dim=axis)


def tile(x, repeat):
    return torch.Tensor.repeat(x, *repeat if isinstance(repeat, tuple) else [repeat])


def norm(x, ord=None, axis=None, keepdims=False, dim=None):
    axis = parse_dim(axis, dim)
    return torch.norm(x, p=ord, dim=axis, keepdim=keepdims)


def min(x, axis=None, other=None, out=None, dim=None):
    axis = parse_dim(axis, dim)
    if other is None:
        axis = axis if axis is not None else 0
        val, _ = torch.min(x, dim=axis, out=out)
        return val
    return torch.min(x, other=other, out=out)


def max(x, axis=None, other=None, out=None, dim=None):
    axis = parse_dim(axis, dim)
    if other is None:
        axis = axis if axis is not None else 0
        val, _ = torch.max(x, dim=axis, out=out)
        return val
    return torch.max(x, other=other, out=out)


def unsqueeze(x, axis=0, dim=0):
    axis = parse_dim(axis, dim)
    return x.unsqueeze(axis)


def where(cond, a, b, *args, **kwargs):
    was_bool = False
    if a.dtype == dtype.bool:
        a = a.to(dtype.int32)
        was_bool = True
    if b.dtype == dtype.bool:
        b = b.to(dtype.int32)
        was_bool = True
    res = torch.where(cond, a, b)
    return res.to(dtype.bool) if was_bool else res


def argmax(x, axis=None, dim=None, *args, **kwargs):
    axis = parse_dim(axis, dim)
    # where not implemented for bool in
    axis = axis if axis is not None else 0
    was_bool = False
    if x.dtype == dtype.bool:
        x = x.to(dtype.int32)
        was_bool = True
    res = torch.argmax(x, dim=axis, *args, **kwargs)
    return res.to(dtype.bool) if was_bool else res


def sqrt(x, *args, **kwargs):
    # where not implemented for bool in
    was_bool = False

    if x.dtype in (dtype.int64, dtype.int32):
        _x = x.to(dtype.float32)
        was_bool = True
    else:
        _x = x
    res = torch.sqrt(_x, *args, **kwargs)
    return res.to(x.dtype) if was_bool else res


def logical_or(a, b):
    return a + b


def logical_and(a, b):
    return a & b



from judo import functions
from judo.judo_backend import Backend


AVAILABLE_FUNCTIONS = set(
    [
        "argmax",
        "hash_numpy",
        "hash_tensor",
        "concatenate",
        "stack",
        "clip",
        "repeat",
        "min",
        "max",
        "norm",
        "unsqueeze",
        "where",
        "sqrt",
        "tile",
        "logical_or",
        "logical_and",
    ]
    + list(functions.batching.AVAILABLE_FUNCTIONS)
    + list(functions.fractalai.AVAILABLE_FUNCTIONS)
    + list(functions.images.AVAILABLE_FUNCTIONS)
    + list(functions.notebook.AVAILABLE_FUNCTIONS),
)


class MetaAPI(type):
    def __getattr__(self, item):
        return self.get_function(name=item)

    @staticmethod
    def get_function(name):
        if name in functions.fractalai.AVAILABLE_FUNCTIONS:
            return getattr(functions.fractalai, name)
        elif name in functions.batching.AVAILABLE_FUNCTIONS:
            return getattr(functions.batching, name)
        elif name in functions.images.AVAILABLE_FUNCTIONS:
            return getattr(functions.images, name)
        elif name in functions.notebook.AVAILABLE_FUNCTIONS:
            return getattr(functions.notebook, name)
        elif Backend.is_numpy():
            backend = functions.numpy
        else:
            backend = functions.pytorch
        return getattr(backend, name)


class API(metaclass=MetaAPI):
    pass



from typing import Callable

import numpy

from judo.data_types import dtype
from judo.judo_backend import Backend, torch
from judo.typing import Tensor


def to_backend_wrap(func) -> Callable[..., Tensor]:
    def wrapped(*args, **kwargs):
        return to_backend(func(*args, **kwargs))

    return wrapped


def is_floating_tensor(x):
    if isinstance(x, numpy.ndarray):
        return isinstance(x.flat[0], (numpy.floating, float))
    elif isinstance(x, torch.Tensor):
        return x.is_floating_point()
    return False


def is_floating_object(x):
    return (
        is_floating_tensor(x)
        or isinstance(x, float)
        or (isinstance(x, (tuple, list)) and isinstance(x[0], float))
    )


def _new_torch_tensor_copy(x, dtype=None, device=None, requires_grad=None, pin_memory=False):
    # TODO(guillemdb): benchmark different alternatives and implement the fastest
    requires_grad = requires_grad and is_floating_object(x)
    return torch.tensor(
        x,
        dtype=dtype,
        device=device,
        requires_grad=requires_grad,
        pin_memory=pin_memory,
    )


def _new_torch_tensor_avoid_copy(x, dtype=None, device=None, requires_grad=None):
    if isinstance(x, torch.Tensor):
        y = x.to(dtype=dtype, device=device)
    # elif isinstance(x, numpy.ndarray):
    #    y = torch.from_numpy(x)
    #    y = (y if (device is None and dtype is None) else
    #    y.to(device=device, dtype=dtype, copy=False))
    elif isinstance(x, torch.autograd.grad_mode.no_grad):
        return x
    else:
        y = torch.as_tensor(x, dtype=dtype, device=device)
    return (
        y
        if requires_grad is not None or not requires_grad or not is_floating_tensor(y)
        else y.requires_grad_()
    )


def new_torch_tensor(x, dtype=None, device=None, requires_grad=None, copy=False, pin_memory=False):
    if copy:
        return _new_torch_tensor_copy(
            x,
            dtype=dtype,
            device=device,
            requires_grad=requires_grad,
            pin_memory=pin_memory,
        )
    return _new_torch_tensor_avoid_copy(x, dtype=dtype, device=device, requires_grad=requires_grad)


def new_numpy_array(x, dtype=None, copy=False, allow_objects: bool = False):
    try:
        return numpy.array(x, dtype=dtype, copy=copy)
    except Exception as e:
        if allow_objects:
            return numpy.array(tuple(x), dtype=object, copy=copy)
        raise e


def new_backend_tensor(
    x,
    dtype=None,
    device=None,
    requires_grad: bool = None,
    copy: bool = False,
    pin_memory: bool = False,
):
    kwargs = update_with_backend_values(
        requires_grad=requires_grad,
        device=device,
        copy=copy,
        dtype=dtype,
    )
    if Backend.is_numpy():
        return new_numpy_array(x, dtype=kwargs.get("dtype"), copy=kwargs.get("copy"))
    elif Backend.is_torch():
        return new_torch_tensor(
            x,
            dtype=kwargs.get("dtype"),
            copy=kwargs.get("copy"),
            device=kwargs.get("device"),
            requires_grad=kwargs.get("requires_grad"),
            pin_memory=pin_memory,
        )


def copy_torch(x: torch.Tensor, requires_grad, device=None) -> torch.Tensor:
    grad = requires_grad if requires_grad is not None else Backend.requires_grad()
    device = torch.device(device) if device is not None else Backend.get_device()
    new_tensor = x.clone()
    if device is not None and new_tensor.device == device:
        new_tensor = new_tensor.to(device)
    if grad is not None and not grad:
        new_tensor = new_tensor.detach()
    return new_tensor


def astype(x, dtype):
    funcs = {
        "numpy": lambda x: x.astype(dtype),
        "torch": lambda x: x.to(dtype),
    }
    return Backend.execute(x, funcs)


def as_tensor(x, *args, **kwargs):
    funcs = {
        "numpy": lambda x: numpy.ascontiguousarray(x, *args, **kwargs),
        "torch": lambda x: torch.as_tensor(x, *args, **kwargs),
    }
    return Backend.execute(x, funcs)


def to_numpy(x, copy=False):
    if isinstance(x, numpy.ndarray):
        return x.copy() if copy else x
    elif isinstance(x, torch.Tensor):
        return x.cpu().detach().numpy()
    else:
        return new_numpy_array(x, copy=copy)


def _assign_device_and_grad(x, device, requires_grad):
    if device is not None:
        x = x.to(device=device)
    if requires_grad is not None and x.requires_grad != requires_grad and is_floating_tensor(x):
        x = x.requires_grad_(requires_grad)
    return x


def to_torch(x, requires_grad: bool = None, device: str = None, copy: bool = False):
    use_grad = requires_grad if requires_grad is not None else Backend.requires_grad()
    device = device if device is not None else Backend.get_device()
    if isinstance(x, torch.Tensor):
        return (
            copy_torch(x, device=device, requires_grad=use_grad)
            if copy
            else _assign_device_and_grad(x, device=device, requires_grad=use_grad)
        )
    return new_torch_tensor(x, device=device, copy=copy)


def update_with_backend_values(**kwargs):
    user_kwargs = {k: v for k, v in kwargs.items() if v is not None}
    backend_kwargs = Backend.get_backend_state()
    backend_kwargs.update(user_kwargs)
    return backend_kwargs


def to_backend(
    x: "Tensor",
    requires_grad: bool = None,
    device: str = None,
    copy: bool = None,
) -> Tensor:
    kwargs = update_with_backend_values(requires_grad=requires_grad, device=device, copy=copy)
    if Backend.is_numpy():
        return to_numpy(x, kwargs.get("copy"))
    return to_torch(
        x,
        requires_grad=kwargs.get("requires_grad"),
        device=kwargs.get("device"),
        copy=kwargs.get("copy"),
    )


def match_backend(x, other) -> Tensor:
    if isinstance(x, numpy.ndarray):
        return to_numpy(other)
    elif isinstance(x, torch.Tensor):
        return torch.tensor(other)


class JudoTensor:
    def __new__(
        cls,
        x,
        dtype=None,
        device: str = None,
        requires_grad: bool = None,
        copy: bool = None,
        pin_memory: bool = False,
    ) -> Tensor:
        return new_backend_tensor(
            x,
            requires_grad=requires_grad,
            device=device,
            copy=copy,
            pin_memory=pin_memory,
            dtype=dtype,
        )


tensor = JudoTensor
array = JudoTensor


def copy(x, requires_grad: bool = None):
    if x is None:
        return
    if not dtype.is_tensor(x):
        x = JudoTensor(x)

    funcs = {
        "numpy": lambda x: x.copy(),
        "torch": lambda x: copy_torch(x, requires_grad),
    }
    return Backend.execute(x, funcs)



import numpy
import pytest
import torch

import judo
from judo import Backend


@pytest.fixture()
def backend():
    return Backend


class TestDataTypes:
    def test_bool(self, backend):
        backend.set_backend("numpy")
        assert judo.bool == numpy.bool_, judo.bool
        backend.set_backend("torch")
        assert judo.bool == torch.bool



import pytest

import judo
from judo import Backend


@pytest.fixture()
def backend():
    return Backend


class TestAPI:
    def test_from_judo(self, backend):
        x = judo.zeros((10, 10))
        assert judo.sqrt(x).sum() == 0



import copy
from typing import Dict, Generator, Iterable, List, Optional, Tuple, Union

import numpy

import judo
from judo.functions.hashing import hasher
from judo.judo_tensor import tensor
from judo.typing import StateDict, Tensor


class States:
    """
    Data structure that handles teh data that defines a population of agents.

    Each population attribute will be stored as a tensor with its first dimension \
    (batch size) representing each agent.

    In order to define a tensor attribute, a `state_dict` dictionary needs to be \
    specified using the following structure::

        state_dict = {"name_1": {"size": tuple([1]),
                                 "dtype": numpy.float32,
                                },
                     }

    Where tuple is a tuple indicating the shape of the desired tensor. The \
    created arrays will accessible the ``name_1`` attribute of the class, or \
    indexing the class with ``states["name_1"]``.

    If ``size`` is not defined the attribute will be considered a vector of \
    length `batch_size`.


    Args:
        batch_size: The number of items in the first dimension of the tensors.
        state_dict: Dictionary defining the attributes of the tensors.
        **kwargs: Data can be directly specified as keyword arguments.

    """

    def __init__(self, batch_size: int, state_dict: Optional[StateDict] = None, **kwargs):
        """
        Initialize a :class:`States`.

        Args:
             batch_size: The number of items in the first dimension of the tensors.
             state_dict: Dictionary defining the attributes of the tensors.
             **kwargs: The name-tensor pairs can also be specified as kwargs.

        """
        attr_dict = self.params_to_arrays(state_dict, batch_size) if state_dict is not None else {}
        attr_dict.update(kwargs)
        self._tensor_names = [k for k, v in attr_dict.items() if judo.is_tensor(v)]
        self._names = list(attr_dict.keys())
        self._attr_dict = attr_dict
        self.update(**self._attr_dict)
        self._batch_size = batch_size

    def __len__(self) -> int:
        """Length is equal to batch_size."""
        return self._batch_size

    def __getitem__(self, item: Union[str, int]) -> Union[Tensor, List[Tensor], "States"]:
        """
        Query an attribute of the class as if it was a dictionary.

        Args:
            item: Name of the attribute to be selected.

        Returns:
            The corresponding item.

        """
        if isinstance(item, str):
            try:
                return getattr(self, item)
            except AttributeError:
                raise TypeError("Tried to get a non existing attribute with key {}".format(item))
        elif judo.is_int(item):
            return self._ix(item)
        else:
            raise TypeError(
                "item must be an instance of str, got {} of type {} instead".format(
                    item,
                    type(item),
                ),
            )

    def _ix(self, index: int):
        # TODO(guillemdb): Allow slicing
        data = {
            k: judo.unsqueeze(v[index], 0) if judo.is_tensor(v) else v for k, v in self.items()
        }
        return self.__class__(batch_size=1, **data)

    def __setitem__(self, key, value: Union[Tuple, List, Tensor]):
        """
        Allow the class to set its attributes as if it was a dict.

        Args:
            key: Attribute to be set.
            value: Value of the target attribute.

        Returns:
            None.

        """
        if key not in self._names:
            self._names.append(key)
        self.update(**{key: value})

    def __repr__(self):
        string = "{} with {} walkers\n".format(self.__class__.__name__, self.n)
        for k, v in self.items():
            shape = v.shape if hasattr(v, "shape") else None
            new_str = "{}: {} {}\n".format(k, type(v), shape)
            string += new_str
        return string

    def __hash__(self) -> int:
        return hasher.hash_state(self)

    def group_hash(self, name: str) -> int:
        """Return a unique id for a given attribute."""
        return hasher.hash_tensor(self[name])

    def hash_walkers(self, name: str) -> List[int]:
        """Return a unique id for each walker attribute."""
        return hasher.hash_walkers(self[name])

    @staticmethod
    def merge_states(states: Iterable["States"]) -> "States":
        """
        Combine different states containing the same kind of data into a single \
        :class:`State` with batch size equal to the sum of all the state batch \
        sizes.

        Args:
            states: Iterable returning :class:`States` with the same attributes.

        Returns:
            :class:`States` containing the combined data of the input values.

        """

        def merge_one_name(states_list, name):
            vals = []
            for state in states_list:
                data = state[name]
                # Attributes that are not tensors are not stacked.
                if not judo.is_tensor(data):
                    return data
                state_len = len(state)
                if len(data.shape) == 0 and state_len == 1:
                    # Name is scalar vector. Data is typing.Scalar value. Transform to array first
                    value = tensor([data]).flatten()
                elif len(data.shape) == 1 and state_len == 1:
                    if data.shape[0] == 1:
                        # Name is typing.Scalar vector. Data already transformed to an array
                        value = data
                    else:
                        # Name is a matrix of vectors. Data needs an additional dimension
                        value = tensor([data])
                elif len(data.shape) == 1 and state_len > 1:
                    # Name is a typing.Scalar vector. Data already has is a one dimensional array
                    value = data
                elif (
                    len(data.shape) > 1
                    and state_len > 1
                    or len(data.shape) > 1
                    and len(state) == 1
                ):
                    # Name is a matrix of vectors. Data has the correct shape
                    value = data
                else:
                    raise ValueError(
                        "Could not infer data concatenation for attribute %s  with shape %s"
                        % (name, data.shape),
                    )
                vals.append(value)
            return judo.concatenate(vals)

        # Assumes all states have the same names.
        data = {name: merge_one_name(states, name) for name in states[0]._names}
        batch_size = sum(s.n for s in states)
        return states[0].__class__(batch_size=batch_size, **data)

    @property
    def n(self) -> int:
        """Return the batch_size of the vectors, which is equivalent to the number of walkers."""
        return self._batch_size

    def get(self, key: str, default=None):
        """
        Get an attribute by key and return the default value if it does not exist.

        Args:
            key: Attribute to be recovered.
            default: Value returned in case the attribute is not part of state.

        Returns:
            Target attribute if found in the instance, otherwise returns the
             default value.

        """
        if key not in self.keys():
            return default
        return self[key]

    def keys(self) -> Generator:
        """Return a generator for the attribute names of the stored data."""
        return (name for name in self._names if not name.startswith("_"))

    def vals(self) -> Generator:
        """Return a generator for the values of the stored data."""
        return (self[name] for name in self._names if not name.startswith("_"))

    def items(self) -> Generator:
        """Return a generator for the attribute names and the values of the stored data."""
        return ((name, self[name]) for name in self._names if not name.startswith("_"))

    def itervals(self):
        """
        Iterate the states attributes by walker.

        Returns:
            Tuple containing all the names of the attributes, and the values that
            correspond to a given walker.

        """
        if self.n <= 1:
            yield self.vals()
            raise StopIteration
        for i in range(self.n):
            yield tuple([v[i] for v in self.vals()])

    def iteritems(self):
        """
        Iterate the states attributes by walker.

        Returns:
            Tuple containing all the names of the attributes, and the values that
            correspond to a given walker.

        """
        if self.n < 1:
            return self.vals()
        for i in range(self.n):
            values = (v[i] if judo.is_tensor(v) else v for v in self.vals())
            yield tuple(self._names), tuple(values)

    def split_states(self, n_chunks: int) -> Generator["States", None, None]:
        """
        Return a generator for n_chunks different states, where each one \
        contain only the data corresponding to one walker.
        """

        def get_chunck_size(state, start, end):
            for name in state._names:
                attr = state[name]
                if judo.is_tensor(attr):
                    return len(attr[start:end])
            return int(numpy.ceil(self.n / n_chunks))

        for start, end in judo.similiar_chunks_indexes(self.n, n_chunks):
            chunk_size = get_chunck_size(self, start, end)
            data = {k: val[start:end] if judo.is_tensor(val) else val for k, val in self.items()}
            new_state = self.__class__(batch_size=chunk_size, **data)
            yield new_state

    def update(self, other: "States" = None, **kwargs):
        """
        Modify the data stored in the States instance.

        Existing attributes will be updated, and new attributes will be created if needed.

        Args:
            other: State class that will be copied upon update.
            **kwargs: It is possible to specify the update as key value attributes, \
                     where key is the name of the attribute to be updated, and value \
                      is the new value for the attribute.
        """

        def update_or_set_attributes(attrs: Union[dict, States]):
            for name, val in attrs.items():
                setattr(self, name, val)

        if other is not None:
            update_or_set_attributes(other)
        if kwargs:
            update_or_set_attributes(kwargs)

    def get_params_dict(self) -> StateDict:
        """Return a dictionary describing the data stored in the :class:`States`."""
        return {
            k: {"shape": v.shape, "dtype": v.dtype}
            for k, v in self.__dict__.items()
            if judo.is_tensor(v)
        }

    def copy(self) -> "States":
        """Crete a copy of the current instance."""
        param_dict = {
            str(name): judo.copy(val) if judo.is_tensor(val) else copy.deepcopy(val)
            for name, val in self.items()
        }
        return States(batch_size=self.n, **param_dict)

    @staticmethod
    def params_to_arrays(param_dict: StateDict, n_walkers: int) -> Dict[str, Tensor]:
        """
        Create a dictionary containing the arrays specified by param_dict.

        Args:
            param_dict: Dictionary defining the attributes of the tensors.
            n_walkers: Number items in the first dimension of the data tensors.

        Returns:
              Dictionary with the same keys as param_dict, containing arrays specified \
              by `param_dict` values.

        """
        tensor_dict = {}
        for key, val in param_dict.items():
            # Shape already includes the number of walkers. Remove walkers axis to create size.
            shape = val.get("shape")
            if shape is None:
                val_size = val.get("size")
            elif len(shape) > 1:
                val_size = shape[1:]
            else:
                val_size = val.get("size")
            # Create appropriate shapes with current state's number of walkers.
            sizes = n_walkers if val_size is None else tuple([n_walkers]) + val_size
            if "size" in val:
                del val["size"]
            if "shape" in val:
                del val["shape"]
            tensor_dict[key] = judo.zeros(sizes, **val)
        return tensor_dict



import judo
from judo.typing import Tensor


AVAILABLE_FUNCTIONS = {"resize_image"}


def resize_image(frame: Tensor, width: int, height: int, mode: str = "RGB") -> Tensor:
    """
    Use PIL to resize an RGB frame to an specified height and width.

    Args:
        frame: Target numpy array representing the image that will be resized.
        width: Width of the resized image.
        height: Height of the resized image.
        mode: Passed to Image.convert.

    Returns:
        The resized frame that matches the provided width and height.

    """
    from PIL import Image

    frame = judo.to_numpy(frame)
    with judo.Backend.use_backend(name="numpy"):
        frame = Image.fromarray(frame)
        frame = judo.to_numpy(frame.convert(mode).resize(size=(width, height)))
    return judo.tensor(frame)



import copy
from typing import Any, Dict, Generator, List, Set, Tuple, Union

import networkx as nx

import judo
from judo.data_structures import States
from judo.functions import hasher, random_state
from judo.judo_backend import Backend
from judo.typing import NamesData, NodeData, NodeDataGenerator, NodeId, Tensor


def to_node_id(x):
    if Backend.is_numpy():
        return str(x) if Backend.use_true_hash() else int(x)
    elif Backend.is_torch():
        return int(x)


class BaseTree:
    """Data structure in charge of storing the history of visited states of an algorithm run."""

    def __init__(self, root_id):
        """
        Initialize a :class:`BaseTree`.

        Args:
            root_id: The node id of the root node.

        """
        self.root_id = root_id

    def __call__(self, *args, **kwargs) -> "BaseTree":
        """
        Return the current instance of :class:`BaseTree`.

        This is used to avoid defining a ``tree_callable `` as \
        ``lambda: tree_instance`` when initializing a :class:`Swarm`. If the \
        :class:`BaseTree` needs is passed to a remote process, you may need \
        to write custom serialization for it, or resort to creating an appropriate \
        ``tree_callable``.
        """
        return self

    def add_states(
        self,
        parent_ids: List[int],
        n_iter: int = None,
        **kwargs,
    ) -> None:
        """
        Update the history of the tree adding the necessary data to recreate a \
        the trajectories sampled by the :class:`Swarm`.

        Args:
            parent_ids: List of states hashes representing the parent nodes of \
                        the current states.
            n_iter: Number of iteration of the algorithm when the data was sampled.
            kwargs: Keyword arguments representing different :class:`States` instances.

        Returns:
            None

        """
        pass

    def reset(self, *args, **kwargs) -> None:
        """
        Delete all the data currently stored and reset the internal state of \
        the tree.
        """
        pass

    def prune_tree(self, *args, **kwargs) -> None:
        """Remove branches of the tree."""
        pass


class NetworkxTree(BaseTree):
    """
    It is a tree data structure that stores the paths followed by the walkers \
    using a networkx DiGraph to keep track of the states relationships.
    """

    DEFAULT_NEXT_PREFIX = "next_"
    DEFAULT_NODE_DATA = (
        "states",
        "observs",
        "compas_clone",
        "processed_rewards",
        "virtual_rewards",
        "distances",
        "clone_probs",
        "will_clone",
        "in_bounds",
        "dt",
        "rewards",
        "oobs",
        "terminals",
        "cum_rewards",
    )

    DEFAULT_EDGE_DATA = ("actions", "dt")

    def __init__(
        self,
        names: NamesData = None,
        prune: bool = False,
        root_id: NodeId = 0,
        node_names: NamesData = DEFAULT_NODE_DATA,
        edge_names: NamesData = DEFAULT_EDGE_DATA,
        next_prefix: str = DEFAULT_NEXT_PREFIX,
    ):
        """
        Initialize a :class:`HistoryTree`.

        Args:
            names: Names of the data attributes that will be extracted stored \
                   in the graph data. The data generators will return the data
                   as a tuple of arrays, ordered according to ``names``.
            prune: If ``True`` the tree will be pruned after every iteration to \
                  remove the branches that have stopped its expansion process.
            root_id: The node id of the root node.
            next_prefix: Prefix used to refer to data extracted from the next \
                        node when parsing a data generator. For example: \
                        "next_observs" will reference the observation of the \
                        next node.
            node_names: Names of the data attributes of the :class:`States` that \
                       will be stored as node attributes in the internal graph.
            edge_names: Names of the data attributes of the :class:`States` that \
                       will be stored as edge attributes in the internal graph.
        """
        super(NetworkxTree, self).__init__(root_id=to_node_id(root_id))
        self.next_prefix = next_prefix
        self.names = names if names is not None else []
        self.data: nx.DiGraph = nx.DiGraph()
        self._node_count = 0
        self.leafs = set()
        self.last_added = set()
        # Node and edge names are parsed with no prefix
        self.prune = prune
        self.node_names = [
            name.lstrip(next_prefix)
            for name in self.names
            if name.lstrip(next_prefix) in node_names
        ]
        self.edge_names = [
            name.lstrip(next_prefix)
            for name in self.names
            if name.lstrip(next_prefix) in edge_names
        ]

    def __len__(self) -> int:
        return self._node_count

    def __repr__(self) -> str:
        string = "%s \n" "Nodes: %s Leafs: %s " % (
            self.__class__.__name__,
            self._node_count,
            len(self.leafs),
        )
        return string

    def reset(self, root_id: NodeId = None, **kwargs) -> None:
        """
        Delete all the data currently stored and reset the internal state of \
        the tree and reset the root node with the provided data.
        """
        node_data = self._extract_data_from_states(index=0, only_node_data=True, **kwargs)
        self.reset_graph(root_id=to_node_id(root_id), node_data=node_data, epoch=-1)

    def get_states_ids(self, **kwargs) -> Tuple[Tensor, States]:
        raise NotImplementedError()

    def add_states(self, parent_ids: List[NodeId], n_iter: int = None, **kwargs):
        """
        Update the history of the tree adding the necessary data to recreate a \
        the trajectories sampled by the :class:`Swarm`.

        Args:
            parent_ids: List of states hashes representing the parent nodes of \
                        the current states.
            n_iter: Number of iteration of the algorithm when the data was sampled.
            kwargs: Keyword arguments representing different :class:`States` instances.

        Returns:
            None

        """
        ids, id_states = self.get_states_ids(**kwargs)
        if hasher.uses_true_hash:
            leaf_ids = judo.to_numpy(ids)
        else:
            leaf_ids = hasher.get_array_of_ids(len(id_states))
        parent_ids = judo.to_numpy(parent_ids)
        # Keep track of nodes that are active to make sure that they are not pruned
        self.last_added = set(leaf_ids) | set(parent_ids)
        for i, (leaf, parent) in enumerate(zip(leaf_ids, parent_ids)):
            node_data, edge_data = self._extract_data_from_states(index=i, **kwargs)
            self.append_leaf(
                leaf_id=leaf,
                parent_id=parent,
                node_data=node_data,
                edge_data=edge_data,
                epoch=n_iter,
            )
        if not hasher.uses_true_hash:
            id_states.update(ids=leaf_ids)

    def prune_tree(self, alive_leafs: Set[int]):
        """
        Remove the branches that do not have a walker in their leaves.

        Args:
            alive_leafs: Contains the ids  of the leaf nodes that are being \
                         expanded by the walkers.

        Returns:
            None.

        """
        if self.prune:
            alive_leafs = set(alive_leafs)
            dead_leafs = self.leafs - alive_leafs
            self.prune_dead_branches(dead_leafs=dead_leafs, alive_leafs=alive_leafs)

    def reset_graph(
        self,
        node_data: Dict[str, Any],
        root_id: NodeId = None,
        epoch: int = -1,
    ) -> None:
        """
        Delete all the data currently stored and reset the internal state of \
        the instance.

        Args:
            root_id: The node id of the root node.
            node_data: Dictionary containing the data that will be added to the \
                      new node created. The keys of the dictionary contain the \
                      names of the attributes and its values the corresponding \
                      value added to the node.
            epoch: Epoch of the :class:`Swarm` run when the current node was \
                    generated. Defaults to -1.

        Returns:
            None.

        """
        self.root_id = to_node_id(root_id) if root_id is not None else self.root_id
        self.data: nx.DiGraph = nx.DiGraph()
        self.data.add_node(self.root_id, epoch=epoch, **node_data)
        self._node_count = 1
        self.leafs = {self.root_id}
        self.last_added = set()

    def append_leaf(
        self,
        leaf_id: NodeId,
        parent_id: NodeId,
        node_data: Dict[str, Any],
        edge_data: Dict[str, Any],
        epoch: int = -1,
    ) -> None:
        """
        Add a new state as a leaf node of the tree to keep track of the \
        trajectories of the swarm.

        Args:
            leaf_id: Node id that identifies the new node that will be added \
                    to the tree.
            parent_id: Node id that identifies the parent of the node that will \
                      be added to the tree.
            node_data: Dictionary containing the data that will be added to the \
                      new node created. The keys of the dictionary contain the \
                      names of the attributes and its values the corresponding \
                      value added to the node.
            edge_data: Dictionary containing the data that will be added to the \
                      new edge created from parent_id to leaf_id. The keys of \
                      the dictionary contain the names of the attributes and \
                      its values the corresponding value added to the edge.
            epoch: Epoch of the :class:`Swarm` run when the current node was \
                    generated. Defaults to -1.

        Returns:
            None.

        """
        # Don't add any leaf that creates a cycle in the graph
        if leaf_id not in self.data.nodes:
            if parent_id not in self.data.nodes:
                raise ValueError("Parent not in graph")
            import copy

            self.data.add_node(leaf_id, epoch=epoch, **copy.deepcopy(node_data))
            self.data.add_edge(parent_id, leaf_id, **copy.deepcopy(edge_data))
            # self.data.add_node(leaf_id, epoch=epoch, **node_data)
            # self.data.add_edge(parent_id, leaf_id, **edge_data)
            self.leafs.add(leaf_id)
            self._node_count += 1
            # If parent is no longer a leaf remove it from the list of leafs
            if parent_id in self.leafs:
                self.leafs.remove(parent_id)
        # TODO(guillemdb): Manage graphs that can have cycles
        # elif (
        #    parent_id in self.data.nodes
        #    and leaf_id in self.data.nodes
        #    and (parent_id, leaf_id) not in self.data.edges()
        # ):
        #    self.remove_cycle_and_store_closest_path(parent_id, leaf_id)

    def prune_dead_branches(self, dead_leafs: Set[NodeId], alive_leafs: Set[NodeId]) -> None:
        """
        Prune the orphan leaves that will no longer be used in order to save memory.

        Args:
            dead_leafs: Leaves of the branches that will be removed.
            alive_leafs: Leaves of the branches that will kept being expanded.

        Returns:
            None

        """
        alive_nodes = alive_leafs | self.last_added
        for leaf in set(dead_leafs):
            self.prune_branch(leaf, alive_nodes)
        return

    def prune_branch(self, leaf_id: NodeId, alive_nodes: Set[NodeId]) -> None:
        """
        Recursively prunes a branch that ends in an orphan leaf.

        Args:
            leaf_id: Id that identifies the leaf of the tree. \
                     If ``leaf_id`` is the hash of a walker state ``from_hash`` \
                     needs to be ``True``. Otherwise it refers to a node id of \
                     the leaf node.
            alive_nodes: Nodes of the branches that can be still expanded.

        Returns:
            None

        """
        leaf_id = to_node_id(leaf_id)
        if (
            leaf_id == self.root_id
            or leaf_id in alive_nodes  # Remove only old nodes (inserted for 2+ epochs)
            or len(self.data.out_edges([leaf_id])) > 0  # Has children -> It's not a leaf
        ):
            return
        parent = self.get_parent(leaf_id)
        if parent in alive_nodes:
            return
        self.data.remove_node(leaf_id)
        self._node_count -= 1
        self.leafs.discard(leaf_id)
        if len(self.data.out_edges([parent])) == 0:  # Track if parent becomes a leaf
            self.leafs.add(parent)
        return self.prune_branch(parent, alive_nodes)

    def get_parent(self, node_id: NodeId) -> NodeId:
        """Get the node id of the parent of the target node."""
        try:
            return list(self.data.in_edges(to_node_id(node_id)))[0][0]
        except (KeyError, IndexError):
            raise KeyError("Node %s does not have a parent in the graph" % node_id)

    def get_branch(self, leaf_id: NodeId) -> List[NodeId]:
        """
        Return a list of the node ids of the path between root_node and leaf_id.

        Args:
            leaf_id: Node id of the end of the path that will be returned.

        Returns:
            List containing the nodes found between the root node and ``leaf_id``. \
            It starts with ``self.root_id`` and ends with ``leaf_id``.

        """
        leaf_id = to_node_id(leaf_id)
        nodes = [leaf_id]
        while leaf_id != self.root_id:
            node = self.get_parent(leaf_id)
            nodes.append(node)
            leaf_id = node
        return nodes[::-1]

    def get_path_node_ids(self, leaf_id: NodeId, root_id: NodeId = None) -> List[NodeId]:
        """
        Get the data of the path between ``leaf_id`` and ``root_id``.

        Args:
            leaf_id: Id that identifies the leaf of the tree. \
                     If ``leaf_id`` is the hash of a walker state ``from_hash`` \
                     needs to be ``True``. Otherwise it refers to a node id of \
                     the leaf node.
            root_id: Node id of the root node of the tree.

        Returns:
            List of the node ids between ``root`` and ``leaf_id``.

        """
        leaf_id = to_node_id(leaf_id)
        root = root_id if root_id is not None else self.root_id
        nodes = nx.shortest_path(self.data, root, leaf_id)
        return nodes

    def get_leaf_nodes(self) -> Tuple[NodeId]:
        """Return a list containing all the node ids of the leaves of the tree."""
        return tuple(node for node in self.data.nodes if len(self.data.out_edges([node])) == 0)

    def compose(self, other: "NetworkxTree") -> None:
        """
        Compose the graph of another :class:`BaseNetworkxTree` with the graph of \
        the current instance.

        Composition is the simple union of the node sets and edge sets.
        The node sets of ``self.data`` and ``other.data`` do not need to be \
        disjoint.

        The data in ``other`` takes precedence over the data in the current instance.

        Args:
            other: Instance of :class:`BaseNetworkxTree` that will be composed \
                  with the current instance.

        Returns:
            None

        """
        self.data = nx.compose(self.data, other.data)

    def _extract_data_from_states(
        self,
        index,
        env_states,
        model_states,
        walkers_states,
        only_node_data: bool = False,
    ):
        def get_name_from_states(name):
            if hasattr(env_states, name):
                data = getattr(env_states, name)[index]
            elif hasattr(model_states, name):
                data = getattr(model_states, name)[index]
            elif hasattr(walkers_states, name):
                data = getattr(walkers_states, name)[index]
            else:
                raise ValueError("Data not present in states: %s" % name)
            return copy.deepcopy(data)

        node_data = {name: get_name_from_states(name) for name in self.node_names}
        if only_node_data:
            return node_data
        edge_data = {name: get_name_from_states(name) for name in self.edge_names}
        return node_data, edge_data


class HistoryTree(NetworkxTree):
    """
    Tree data structure that keeps track of the visited states.

    It allows to save the :class:`Swarm` data after every iteration and methods to \
    recover the sampled data after the algorithm run.

    The data that will be stored in the graph it's defined in the ``names`` parameter.
    For example:

     - If names is ``["observs", "actions"]``, the observations of every visited \
       state will be stored as node attributes, and actions will be stored as edge attributes.

     - If names if ``["observs", "actions", "next_observs"]`` the same data will be stored,
       but when the data generator methods are called the observation corresponding \
       to the next state will also be returned.

     The attribute ``names`` also defines the order of the data returned by the generator.

     As long as the data is stored in the graph (passing a valid ``names`` list at \
     initialization, the order of the data can be redefined passing the ``names`` \
     parameter to the generator method.

     For example, if the ``names`` passed at initialization is ``["states", "rewards"]``, \
     you can call the generator methods with ``names=["rewards", "states", "next_states"]`` \
     and the returned data will be a tuple containing (rewards, states, next_states).

    """

    def __init__(
        self,
        names: NamesData = None,
        prune: bool = False,
        root_id: NodeId = 0,
        node_names: NamesData = NetworkxTree.DEFAULT_NODE_DATA,
        edge_names: NamesData = NetworkxTree.DEFAULT_EDGE_DATA,
        next_prefix: str = NetworkxTree.DEFAULT_NEXT_PREFIX,
    ):
        """
        Initialize a :class:`HistoryTree`.

        Args:
            names: Names of the data attributes that will be extracted stored \
                   in the graph data. The data generators will return the data
                   as a tuple of arrays, ordered according to ``names``.
            prune: If ``True`` the tree will be pruned after every iteration to \
                  remove the branches that have stopped its expansion process.
            root_id: The node id of the root node.
            next_prefix: Prefix used to refer to data extracted from the next \
                        node when parsing a data generator. For example: \
                        "next_observs" will reference the observation of the \
                        next node.
            node_names: Names of the data attributes of the :class:`States` that \
                       will be stored as node attributes in the internal graph.
            edge_names: Names of the data attributes of the :class:`States` that \
                       will be stored as edge attributes in the internal graph.
        """
        super(HistoryTree, self).__init__(
            names=names,
            prune=prune,
            root_id=root_id,
            next_prefix=next_prefix,
            node_names=node_names,
            edge_names=edge_names,
        )

    def _one_node_tuples(self, node, next_node, return_children) -> NodeData:
        node_data, edge_data = self.data.nodes[node], self.data.edges[(node, next_node)]
        if return_children:
            return node_data, edge_data, self.data.nodes[next_node]
        else:
            return node_data, edge_data

    def _extract_one_node_data(self, data: NodeData, names: NamesData) -> Tuple:
        """
        Transform the one tuple of dictionaries returned by the data generators \
        into a tuple of data values where each value corresponds to a name in names.
        """

        def get_item(data, name):
            node_data, edge_data, *next_node_data = data
            if name.startswith(self.next_prefix):
                true_name = name.replace(self.next_prefix, "")
                return next_node_data[0][true_name]
            elif name in node_data:
                return node_data[name]
            elif name in edge_data:
                return edge_data[name]
            raise KeyError("%s not found in data: %s" % (name, data))

        return tuple(get_item(data, name) for name in names)

    def _process_batch(self, batch_data: List[tuple]) -> Tuple[Tensor, ...]:
        """
        Preprocess the list of tuples representing a batched group of elements \
        and return a tuple of arrays representing the batched values for every \
        data attribute.
        """
        unpacked = zip(*batch_data)
        return tuple(judo.as_tensor(val) for val in unpacked)

    def _generate_batches(
        self,
        generator: NodeDataGenerator,
        names: NamesData,
        batch_size: int = None,
    ) -> Generator[Tuple, None, None]:
        """Return batches of processed data represented as tuples of arrays."""
        returned = []
        batch_counter = 0
        for i, data in enumerate(generator):
            batch_counter += 1
            extracted_data = self._extract_one_node_data(data, names)
            if batch_size is None:  # Do not batch data
                yield extracted_data
                continue
            returned.append(extracted_data)
            if batch_counter == batch_size:
                batch_counter = 0
                yield self._process_batch(returned)
                returned = []
            elif batch_size > 0 and i == len(self) - batch_size:
                break
        # If a batch size less than 1 is provided return all the data as a single batch
        if batch_size is not None and batch_size <= 1:
            yield self._process_batch(returned)

    def _validate_names(self, names):
        if names is None:
            return self.names
        for name in names:
            name = (
                name[len(self.next_prefix) :]  # noqa: E203
                if name.startswith(self.next_prefix)
                else name
            )
            if name not in self.names:
                raise KeyError(
                    "Data corresponding to name %s "
                    "not present in self.names: %s for element %s"
                    % (name, self.names, name[len(self.next_prefix) :]),  # noqa: E203
                )
        return names

    def path_data_generator(
        self,
        node_ids: Union[Tuple[NodeId], Set[NodeId], List[NodeId]],
        return_children: bool = False,
    ) -> NodeDataGenerator:
        """
        Return a generator that returns the data corresponding to a path.

        Each value yielded corresponds to one element of the path. The edge data \
        will be assigned to the parent node of the  edge. This means the edge \
        data contains the data associated with the transition to a child.

        The data corresponding to the last node of the path will not be returned.

        Args:
            node_ids: Node ids of a path ordered. The first node of the path \
                     corresponds to the first element of ``node_ids``.
            return_children: If ``True`` the data corresponding to the child of \
                            each node in the path will also be returned.

        Yields:
            tuple of dictionaries containing the data of each node and edge of \
            the path, following the order of ``node_ids``.

            If ``return_children`` is ``False`` it will yield (node_data, \
            edge_data).
            If ``return_children`` is ``True`` it will yield (node_data, \
            edge_data, next_node_data).

        """
        for (i, node) in enumerate(node_ids[:-1]):
            next_node = node_ids[i + 1]
            yield self._one_node_tuples(node, next_node, return_children)

    def random_nodes_generator(self, return_children: bool = False) -> NodeDataGenerator:
        """
        Return a generator that yields the data of all the nodes sampling them at random.

        Each value yielded corresponds to the data associated with one node and the \
        edge connecting to one of its children. The edge data will be assigned \
        to the parent node of the edge. This means the edge data of each node \
        contains the data associated with the transition to a child.

        Args:
            return_children: If ``True`` the data corresponding to the child of \
                            each node in the path will also be returned.

        Yields:
            tuple of dictionaries containing the data for each node and edge \
            sampled at random.

            If ``return_children`` is ``False`` it will yield (node_data, \
            edge_data).
            If ``return_children`` is ``True`` it will yield (node_data, \
            edge_data, next_node_data).

        """
        with Backend.use_backend("numpy"):
            permutaion = random_state.permutation(list(self.data.nodes))
        for next_node in permutaion:
            if next_node == self.root_id:
                continue
            node = self.get_parent(next_node)
            yield self._one_node_tuples(node, next_node, return_children)

    def iterate_path_data(
        self,
        node_ids: Union[Tuple[NodeId], List[NodeId]],
        batch_size: int = None,
        names: NamesData = None,
    ) -> NodeDataGenerator:
        """
        Return a generator that yields the data of the nodes contained in the provided path.

        Args:
            node_ids: Ids of the nodes of the path that will be sampled. The \
                     nodes must be provided in order, starting with the first \
                     node of the path.
            batch_size: If it is not None, the generator will return batches of \
                        data with the target batch size. If Batch size is less than 1 \
                        it will return a single batch containing all the data.
            names: Names of the data attributes that will be yielded by the generator.

        Returns:
            Generator providing the data corresponding to a path in the internal tree.

        """
        names = self._validate_names(names)
        return_children = any(name.startswith(self.next_prefix) for name in names)
        path_generator = self.path_data_generator(
            node_ids=judo.to_numpy(node_ids),
            return_children=return_children,
        )
        return self._generate_batches(path_generator, names=names, batch_size=batch_size)

    def iterate_nodes_at_random(
        self,
        batch_size: int = None,
        names: NamesData = None,
    ) -> NodeDataGenerator:
        """
        Return a generator that yields the data of the nodes contained in the provided path.

        Args:
            batch_size: If it is not None, the generator will return batches of \
                        data with the target batch size. If Batch size is less than 1 \
                        it will return a single batch containing all the data.
            names: Names of the data attributes that will be yielded by the generator.

        Returns:
            Generator providing the data corresponding to all the nodes of \
            the tree sampled at random.

        """
        names = self._validate_names(names)
        return_children = any(name.startswith(self.next_prefix) for name in names)
        node_generator = self.random_nodes_generator(return_children=return_children)
        return self._generate_batches(node_generator, names=names, batch_size=batch_size)

    def iterate_branch(
        self,
        node_id: NodeId,
        batch_size: int = None,
        names: NamesData = None,
    ) -> NodeDataGenerator:
        """
        Return a generator that yields the data of the nodes contained in the provided path.

        Args:
            node_id: Ids of the last node of the branch that will be sampled. \
                    The first node will be the root node of the tree, and the last \
                    one will be the one with the provided ``node_id``.
            batch_size: If it is not None, the generator will return batches of \
                        data with the target batch size. If Batch size is less than 1 \
                        it will return a single batch containing all the data.
            names: Names of the data attributes that will be yielded by the generator.

        Returns:
            Generator providing the data corresponding to a branch of the internal tree.

        """
        node_id = to_node_id(node_id)
        branch_path = self.get_path_node_ids(leaf_id=node_id)
        return self.iterate_path_data(branch_path, batch_size=batch_size, names=names)



from contextlib import contextmanager
import os

import numpy
import yaml


try:
    import torch
except ImportError:

    class cuda:
        @staticmethod
        def is_available():
            return False

    class torch_random:
        @staticmethod
        def manual_seed(*args, **kwargs):
            return None

    class torch:
        cuda = cuda
        Tensor = numpy.ndarray
        random = torch_random


FALLBACK_DEFAULTS = {
    "backend": "numpy",
    "device": "cpu",
    "requires_grad": None,
    "true_hash": False,
    "copy": False,
}
config_file = os.path.join(os.path.dirname(os.path.abspath(__file__)), "config.yml")


def load_backend_config(filepath=config_file):
    try:
        with open(filepath, "r") as stream:
            config = yaml.safe_load(stream)
    except FileNotFoundError:
        with open(filepath, "w") as file:
            yaml.dump(FALLBACK_DEFAULTS, file)
        config = FALLBACK_DEFAULTS
    backend = config.get("backend", FALLBACK_DEFAULTS["backend"])
    device = config.get("device", FALLBACK_DEFAULTS["device"])
    if device == "auto":
        device = "cuda" if torch.cuda.is_available() else "cpu"
    elif str(device).lower() == "none":
        device = None
    requires_grad = config.get("requires_grad", FALLBACK_DEFAULTS["requires_grad"])
    true_hash = config.get("true_hash", FALLBACK_DEFAULTS["true_hash"])
    copy = config.get("copy", FALLBACK_DEFAULTS["copy"])
    return backend, device, requires_grad, true_hash, copy


def update_backend_config(
    backend: str = None,
    device: str = None,
    requires_grad: bool = None,
    filepath=config_file,
    true_hash: bool = None,
    copy: bool = None,
):
    with open(filepath, "r") as stream:
        config = yaml.safe_load(stream)
        if config is None:
            config = {}
    if backend is not None:
        config["default_backend"] = backend
    if device is not None:
        config["default_device"] = device
    if requires_grad is not None:
        config["requires_grad"] = requires_grad
    if true_hash is not None:
        config["true_hash"] = true_hash
    if copy is not None:
        config["copy"] = copy
    with open(filepath, "w") as outfile:
        yaml.dump(config, outfile)


@contextmanager
def _use_backend(cls, name, device=None, requires_grad=None, copy=None):
    if name is not None:
        cls._check_valid_backend(name)
    curr_state = cls.get_backend_state()
    cls.set_backend(name=name, device=device, requires_grad=requires_grad, copy=copy)
    try:
        yield
    finally:
        cls.set_backend(**curr_state)


@contextmanager
def use_backend(name, device=None, requires_grad=None):
    if name is not None:
        Backend._check_valid_backend(name)
    curr_state = Backend.get_backend_state()
    Backend.set_backend(name=name, device=device, requires_grad=requires_grad)
    try:
        yield
    finally:
        Backend.set_backend(**curr_state)


_backend, _device, _requires_grad, _true_hash, _copy = load_backend_config()
_backend, _device, _requires_grad, _true_hash, _copy = (
    str(_backend),
    str(_device) if _device is not None else None,
    bool(_requires_grad) if _requires_grad is not None else None,
    bool(_true_hash),
    bool(_copy),
)


class Backend:
    AVAILABLE_BACKENDS = ["numpy", "torch"]
    _backend, _device, _requires_grad, _true_hash = _backend, _device, _requires_grad, _true_hash
    _copy = _copy

    @classmethod
    def _check_valid_backend(cls, name):
        if name not in cls.AVAILABLE_BACKENDS:
            raise ValueError(
                "%s not supported. Available backends: %s" % (name, cls.AVAILABLE_BACKENDS),
            )

    @classmethod
    def get_backend_state(cls):
        state = {
            "name": str(cls._backend),
            "device": str(cls._device) if cls._device is not None else None,
            "requires_grad": bool(cls._requires_grad) if cls._requires_grad is not None else None,
            "true_hash": bool(cls._true_hash),
            "copy": bool(cls._copy),
        }
        return state

    @classmethod
    def get_current_backend(cls):
        return cls._backend

    @classmethod
    def get_device(cls):
        if cls._device is not None:
            return str(cls._device)
        return None

    @classmethod
    def requires_grad(cls):
        return cls._requires_grad

    @classmethod
    def use_true_hash(cls):
        return cls._true_hash

    @classmethod
    def set_defaults(
        cls,
        name=None,
        device=None,
        requires_grad: bool = None,
        copy: bool = None,
        true_hash: bool = None,
        set_backend: bool = True,
    ):
        update_backend_config(
            backend=name,
            device=device,
            requires_grad=requires_grad,
            true_hash=true_hash,
            copy=copy,
        )
        if set_backend:
            cls.set_backend(
                name=name,
                device=device,
                requires_grad=requires_grad,
                copy=copy,
                true_hash=true_hash,
            )

    @classmethod
    def set_backend(
        cls,
        name=None,
        device=None,
        requires_grad: bool = None,
        copy: bool = None,
        true_hash: bool = None,
    ):
        if name is not None:
            cls._check_valid_backend(name)
            cls._backend = name
        if device is not None:
            cls._device = device
        if requires_grad is not None:
            cls._requires_grad = requires_grad
        if copy is not None:
            cls._copy = copy
        if true_hash is not None:
            cls._true_hash = true_hash

    @classmethod
    def is_numpy(cls):
        return cls._backend == "numpy"

    @classmethod
    def is_torch(cls):
        return cls._backend == "torch"

    @classmethod
    def can_use_cuda(cls):
        return cls.is_torch() and cls._device != "cpu"

    @classmethod
    def execute(cls, value, funcs):
        backend = cls.get_current_backend()
        return funcs[backend](value)

    @classmethod
    def use_backend(cls, name=None, device=None, requires_grad=None, copy=None):
        return _use_backend(cls, name=name, device=device, requires_grad=requires_grad, copy=copy)

    @classmethod
    def reset_state(cls):
        global _backend, _device, _requires_grad, _true_hash, _copy
        _backend, _device, _requires_grad, _true_hash, _copy = load_backend_config()
        _backend, _device, _requires_grad, _true_hash, _copy = (
            str(_backend),
            str(_device) if _device is not None else None,
            bool(_requires_grad) if _requires_grad is not None else None,
            bool(_true_hash),
            bool(_copy),
        )
        cls._backend, cls._device, cls._requires_grad, cls._true_hash, cls._copy = (
            _backend,
            _device,
            _requires_grad,
            _true_hash,
            _copy,
        )



"""Functionality for running fragile in numpy and pytorch."""
import sys

import numpy

from judo import data_types as _data_types
from judo.data_structures import Bounds, HistoryTree, States
from judo.data_types import DATA_TYPE_NAMES, dtype
from judo.functions.api import API, AVAILABLE_FUNCTIONS
from judo.functions.hashing import hasher
from judo.functions.random import random_state
from judo.judo_backend import Backend, torch
from judo.judo_tensor import (
    array,
    as_tensor,
    astype,
    copy,
    tensor,
    to_backend,
    to_backend_wrap as __backend_wrap,
    to_numpy,
    to_torch,
)


def __base_getattr(name):
    raise AttributeError(f"module '{module.__name__}' has no attribute '{name}'")


def __new_getattr(name):
    if name in DATA_TYPE_NAMES:
        return getattr(_data_types, name)()
    elif name in AVAILABLE_FUNCTIONS:
        return getattr(API, name)
    try:
        return __old_getattr(name)
    except AttributeError as e:
        if Backend.is_numpy():
            val = getattr(numpy, name)
            return __backend_wrap(val) if callable(val) else val
        elif Backend.is_torch():
            val = getattr(torch, name)
            return __backend_wrap(val) if callable(val) else val
        raise e


module = sys.modules[__name__]
__old_getattr = getattr(module, "__getattr__", __base_getattr)
module.__getattr__ = __new_getattr
is_bool = dtype.is_bool
is_int = dtype.is_int
is_float = dtype.is_float
is_tensor = dtype.is_tensor



# import numpy
import pytest

import judo
from judo import tensor
from judo.data_structures.bounds import Bounds


def create_bounds(name):
    if name == "scalars":
        return lambda: Bounds(high=5, low=-5, shape=(3,))
    elif name == "high_array":
        return lambda: Bounds(high=tensor([1, 2, 5], dtype=judo.float), low=-5)
    elif name == "low_array":
        return lambda: Bounds(low=tensor([-1, -5, -3], dtype=judo.float), high=5)
    elif name == "both_array":
        array = tensor([1, 2, 5], dtype=judo.float)
        return lambda: Bounds(high=array, low=-array)
    elif name == "high_list":
        return lambda: Bounds(
            low=tensor([-5, -2, -3], dtype=judo.float), high=[5, 5, 5], dtype=judo.float
        )


bounds_fixture_params = ["scalars", "high_array", "low_array", "both_array", "high_list"]


@pytest.fixture(params=bounds_fixture_params, scope="class")
def bounds_fixture(request) -> Bounds:
    return create_bounds(request.param)()


class TestBounds:
    def test_init(self, bounds_fixture):
        assert bounds_fixture.dtype is not None
        assert judo.is_tensor(bounds_fixture.low)
        assert judo.is_tensor(bounds_fixture.high)
        assert isinstance(bounds_fixture.shape, tuple)
        assert bounds_fixture.low.shape == bounds_fixture.shape
        assert bounds_fixture.high.shape == bounds_fixture.shape

    def test_shape(self, bounds_fixture: Bounds):
        shape = bounds_fixture.shape
        assert isinstance(shape, tuple)
        assert shape == (3,)

    def test_points_in_bounds(self, bounds_fixture):
        points = tensor([[0, 0, 0], [11, 0, 0], [0, 11, 0], [11, 11, 11]])
        res = bounds_fixture.points_in_bounds(points)
        for a, b in zip(res.tolist(), [True, False, False, False]):
            assert a == b

    def test_from_tuples(self):
        tup = ((-1, 2), (-3, 4), (2, 5))
        bounds = Bounds.from_tuples(tup)
        assert (bounds.low == tensor([-1, -3, 2])).all()
        assert (bounds.high == tensor([2, 4, 5])).all()

    def test_from_array(self):
        array = tensor([[0, 0, 0], [11, 0, 0], [0, 11, 0], [11, 11, 11]])
        bounds = Bounds.from_array(array)
        assert (bounds.low == tensor([0, 0, 0])).all()
        assert (bounds.high == tensor([11, 11, 11])).all()
        assert bounds.shape == (3,)

    def test_from_array_with_scale_positive(self):
        array = tensor([[0, 0, 0], [10, 0, 0], [0, 10, 0], [10, 10, 10]], dtype=judo.float)
        bounds = Bounds.from_array(array, scale=1.1)
        assert (bounds.low == tensor([0, 0, 0], dtype=judo.float)).all(), (
            bounds.low,
            array.min(axis=0),
        )
        assert (bounds.high == tensor([11, 11, 11], dtype=judo.float)).all(), (
            bounds.high,
            array.max(axis=0),
        )
        assert bounds.shape == (3,)

        array = tensor([[-10, 0, 0], [-10, 0, 0], [0, -10, 0], [-10, -10, -10]], dtype=judo.float)
        bounds = Bounds.from_array(array, scale=1.1)
        assert (bounds.high == tensor([0, 0, 0], dtype=judo.float)).all(), (
            bounds.high,
            array.max(axis=0),
        )
        assert (bounds.low == tensor([-11, -11, -11], dtype=judo.float)).all(), (
            bounds.low,
            array.min(axis=0),
        )
        assert bounds.shape == (3,)

        array = tensor(
            [[10, 10, 10], [100, 10, 10], [10, 100, 10], [100, 100, 100]], dtype=judo.float
        )
        bounds = Bounds.from_array(array, scale=1.1)
        assert judo.allclose(bounds.low, tensor([9.0, 9.0, 9], dtype=judo.float)), (
            bounds.low,
            array.min(axis=0),
        )
        assert judo.allclose(bounds.high, tensor([110, 110, 110], dtype=judo.float)), (
            bounds.high,
            array.max(axis=0),
        )
        assert bounds.shape == (3,)

    def test_from_array_with_scale_negative(self):
        # high +, low +, scale > 1
        array = tensor([[-10, 0, 0], [-10, 0, 0], [0, -10, 0], [-10, -10, -10]], dtype=judo.float)
        bounds = Bounds.from_array(array, scale=0.9)
        assert (bounds.high == tensor([0, 0, 0], dtype=judo.float)).all(), (
            bounds.high,
            array.max(axis=0),
        )
        assert (bounds.low == tensor([-9, -9, -9], dtype=judo.float)).all(), (
            bounds.low,
            array.min(axis=0),
        )
        assert bounds.shape == (3,)
        array = tensor([[0, 0, 0], [10, 0, 0], [0, 10, 0], [10, 10, 10]], dtype=judo.float)
        bounds = Bounds.from_array(array, scale=0.9)
        assert (bounds.low == tensor([0, 0, 0], dtype=judo.float)).all(), (bounds, array)
        assert (bounds.high == tensor([9, 9, 9], dtype=judo.float)).all()
        assert bounds.shape == (3,)
        # high +, low +, scale < 1
        array = tensor(
            [[10, 10, 10], [100, 10, 10], [10, 100, 10], [100, 100, 100]], dtype=judo.float
        )
        bounds = Bounds.from_array(array, scale=0.9)
        assert judo.allclose(bounds.low, tensor([9.0, 9.0, 9.0], dtype=judo.float)), (
            bounds.low,
            array.min(axis=0),
        )
        assert judo.allclose(bounds.high, tensor([90, 90, 90], dtype=judo.float)), (
            bounds.high,
            array.max(axis=0),
        )
        assert bounds.shape == (3,)
        # high -, low -, scale > 1
        array = tensor(
            [[-100, -10, -10], [-100, -10, -10], [-10, -100, -10], [-100, -100, -100]],
            dtype=judo.float,
        )
        bounds = Bounds.from_array(array, scale=1.1)
        assert judo.allclose(bounds.high, tensor([-9, -9, -9], dtype=judo.float)), (
            bounds.high,
            array.max(axis=0),
        )
        assert judo.allclose(bounds.low, tensor([-110, -110, -110], dtype=judo.float)), (
            bounds.low,
            array.min(axis=0),
        )
        assert bounds.shape == (3,)
        # high -, low -, scale < 1
        array = tensor(
            [[-100, -10, -10], [-100, -10, -10], [-10, -100, -10], [-100, -100, -100]],
            dtype=judo.float,
        )
        bounds = Bounds.from_array(array, scale=0.9)
        assert judo.allclose(bounds.high, tensor([-11, -11, -11], dtype=judo.float)), (
            bounds.high,
            array.max(axis=0),
        )
        assert judo.allclose(bounds.low, tensor([-90, -90, -90], dtype=judo.float)), (
            bounds.low,
            array.min(axis=0),
        )
        assert bounds.shape == (3,)

    def test_clip(self):
        tup = ((-1, 10), (-3, 4), (2, 5))
        array = tensor([[-10, 0, 0], [11, 0, 0], [0, 11, 0], [11, 11, 11]], dtype=judo.float)
        bounds = Bounds.from_tuples(tup)
        clipped = bounds.clip(array)
        target = tensor(
            [[-1.0, 0.0, 2.0], [10.0, 0.0, 2.0], [0.0, 4.0, 2], [10, 4, 5]], dtype=judo.float
        )
        assert judo.allclose(clipped, target), (clipped.dtype, target.dtype)

    @pytest.mark.parametrize("bounds_fixture", bounds_fixture_params, indirect=True)
    def test_to_tuples(self, bounds_fixture):
        tuples = bounds_fixture.to_tuples()
        assert len(tuples) == 3
        assert min([x[0] for x in tuples]) == -5
        assert max([x[1] for x in tuples]) == 5

    @pytest.mark.parametrize("bounds_fixture", bounds_fixture_params, indirect=True)
    def test_points_in_bounds(self, bounds_fixture):
        zeros = judo.zeros((3, 3))
        assert all(bounds_fixture.points_in_bounds(zeros))
        tens = judo.ones((3, 3)) * 10.0
        res = bounds_fixture.points_in_bounds(tens)
        assert not res.any(), (res, tens)
        tens = tensor([[-10, 0, 1], [0, 0, 0], [10, 10, 10]])
        assert sum(bounds_fixture.points_in_bounds(tens)) == 1

    @pytest.mark.parametrize("bounds_fixture", bounds_fixture_params, indirect=True)
    def test_safe_margin(self, bounds_fixture: Bounds):
        new_bounds = bounds_fixture.safe_margin()
        assert judo.allclose(new_bounds.low, bounds_fixture.low)
        assert judo.allclose(new_bounds.high, bounds_fixture.high)
        low = judo.full_like(bounds_fixture.low, -10)
        new_bounds = bounds_fixture.safe_margin(low=low)
        assert judo.allclose(new_bounds.high, bounds_fixture.high)
        assert judo.allclose(new_bounds.low, low)
        new_bounds = bounds_fixture.safe_margin(low=low, scale=2)
        assert judo.allclose(new_bounds.high, bounds_fixture.high * 2)
        assert judo.allclose(new_bounds.low, low * 2)



from collections.abc import Iterable as _Iterable
from typing import Iterable, Optional, Tuple, Union

import numpy

import judo
from judo.judo_tensor import tensor
from judo.typing import Scalar, Tensor


class Bounds:
    """
    The :class:`Bounds` implements the logic for defining and managing closed intervals, \
    and checking if a numpy array's values are inside a given interval.

    It is used on a numpy array of a target shape.
    """

    def __init__(
        self,
        high: Union[Tensor, Scalar] = numpy.inf,
        low: Union[Tensor, Scalar] = numpy.NINF,
        shape: Optional[tuple] = None,
        dtype: Optional[type] = None,
    ):
        """
        Initialize a :class:`Bounds`.

        Args:
            high: Higher value for the bound interval. If it is an typing_.Scalar \
                  it will be applied to all the coordinates of a target vector. \
                  If it is a vector, the bounds will be checked coordinate-wise. \
                  It defines and closed interval.
            low: Lower value for the bound interval. If it is a typing_.Scalar it \
                 will be applied to all the coordinates of a target vector. \
                 If it is a vector, the bounds will be checked coordinate-wise. \
                 It defines and closed interval.
            shape: Shape of the array that will be bounded. Only needed if `high` and `low` are \
                   vectors and it is used to define the dimensions that will be bounded.
            dtype:  Data type of the array that will be bounded. It can be inferred from `high` \
                    or `low` (the type of `high` takes priority).

        Examples:
            Initializing :class:`Bounds` using  numpy arrays:

            >>> import numpy
            >>> high, low = numpy.ones(3, dtype=float), -1 * numpy.ones(3, dtype=int)
            >>> bounds = Bounds(high=high, low=low)
            >>> print(bounds)
            Bounds shape float64 dtype (3,) low [-1 -1 -1] high [1. 1. 1.]

            Initializing :class:`Bounds` using  typing_.Scalars:

            >>> import numpy
            >>> high, low, shape = 4, 2.1, (5,)
            >>> bounds = Bounds(high=high, low=low, shape=shape)
            >>> print(bounds)
            Bounds shape float64 dtype (5,) low [2.1 2.1 2.1 2.1 2.1] high [4. 4. 4. 4. 4.]

        """
        # Infer shape if not specified
        if shape is None and hasattr(high, "shape"):
            shape = high.shape
        elif shape is None and hasattr(low, "shape"):
            shape = low.shape
        elif shape is None:
            raise TypeError("If shape is None high or low need to have .shape attribute.")
        # High and low will be arrays of target shape
        if not judo.is_tensor(high):
            high = tensor(high) if isinstance(high, _Iterable) else judo.ones(shape) * high
        if not judo.is_tensor(low):
            low = tensor(low) if isinstance(low, _Iterable) else judo.ones(shape) * low
        self.high = judo.astype(high, judo.float)
        self.low = judo.astype(low, judo.float)
        if dtype is not None:
            self.dtype = dtype
        elif hasattr(high, "dtype"):
            self.dtype = high.dtype
        elif hasattr(low, "dtype"):
            self.dtype = low.dtype
        else:
            self.dtype = type(high) if high is not None else type(low)

    def __repr__(self):
        return "{} shape {} dtype {} low {} high {}".format(
            self.__class__.__name__,
            self.dtype,
            self.shape,
            self.low,
            self.high,
        )

    def __len__(self) -> int:
        """Return the number of dimensions of the bounds."""
        return len(self.high)

    @property
    def shape(self) -> Tuple:
        """
        Get the shape of the current bounds.

        Returns:
            tuple containing the shape of `high` and `low`

        """
        return self.high.shape

    @classmethod
    def from_tuples(cls, bounds: Iterable[tuple]) -> "Bounds":
        """
        Instantiate a :class:`Bounds` from a collection of tuples containing \
        the higher and lower bounds for every dimension as a tuple.

        Args:
            bounds: Iterable that returns tuples containing the higher and lower \
                    bound for every dimension of the target bounds.

        Returns:
                :class:`Bounds` instance.

        Examples:
            >>> intervals = ((-1, 1), (-2, 1), (2, 3))
            >>> bounds = Bounds.from_tuples(intervals)
            >>> print(bounds)
            Bounds shape int64 dtype (3,) low [-1 -2  2] high [1 1 3]

        """
        low, high = [], []
        for lo, hi in bounds:
            low.append(lo)
            high.append(hi)
        low, high = tensor(low, dtype=judo.float), tensor(high, dtype=judo.float)
        return Bounds(low=low, high=high)

    @staticmethod
    def get_scaled_intervals(
        low: Union[Tensor, float, int],
        high: Union[Tensor, float, int],
        scale: float,
    ) -> Tuple[Union[Tensor, float], Union[Tensor, float]]:
        """
        Scale the high and low vectors by an scale factor.

        The value of the high and low will be proportional to the maximum and minimum values of \
        the array. Scale defines the proportion to make the bounds bigger and smaller. For \
        example, if scale is 1.1 the higher bound will be 10% higher, and the lower bounds 10% \
        smaller. If scale is 0.9 the higher bound will be 10% lower, and the lower bound 10% \
        higher. If scale is one, `high` and `low` will be equal to the maximum and minimum values \
        of the array.

        Args:
            high: Higher bound to be scaled.
            low: Lower bound to be scaled.
            scale: Value representing the tolerance in percentage from the current maximum and \
            minimum values of the array.

        Returns:
            :class:`Bounds` instance.

        """
        pct = tensor(scale - 1)
        big_scale = 1 + judo.abs(pct)
        small_scale = 1 - judo.abs(pct)
        zero = judo.astype(tensor(0.0), low.dtype)
        if pct > 0:
            xmin_scaled = judo.where(low < zero, low * big_scale, low * small_scale)
            xmax_scaled = judo.where(high < zero, high * small_scale, high * big_scale)
        else:
            xmin_scaled = judo.where(low < zero, low * small_scale, low * small_scale)
            xmax_scaled = judo.where(high < zero, high * big_scale, high * small_scale)
        return xmin_scaled, xmax_scaled

    @classmethod
    def from_array(cls, x: Tensor, scale: float = 1.0) -> "Bounds":
        """
        Instantiate a bounds compatible for bounding the given array. It also allows to set a \
        margin for the high and low values.

        The value of the high and low will be proportional to the maximum and minimum values of \
        the array. Scale defines the proportion to make the bounds bigger and smaller. For \
        example, if scale is 1.1 the higher bound will be 10% higher, and the lower bounds 10% \
        smaller. If scale is 0.9 the higher bound will be 10% lower, and the lower bound 10% \
        higher. If scale is one, `high` and `low` will be equal to the maximum and minimum values \
        of the array.

        Args:
            x: Numpy array used to initialize the bounds.
            scale: Value representing the tolerance in percentage from the current maximum and \
            minimum values of the array.

        Returns:
            :class:`Bounds` instance.

        Examples:
            >>> import numpy
            >>> x = numpy.ones((3, 3))
            >>> x[1:-1, 1:-1] = -5
            >>> bounds = Bounds.from_array(x, scale=1.5)
            >>> print(bounds)
            Bounds shape float64 dtype (3,) low [ 0.5 -7.5  0.5] high [1.5 1.5 1.5]

        """
        xmin, xmax = judo.min(x, axis=0), judo.max(x, axis=0)
        xmin_scaled, xmax_scaled = cls.get_scaled_intervals(xmin, xmax, scale)
        return Bounds(low=xmin_scaled, high=xmax_scaled)

    def clip(self, x: Tensor) -> Tensor:
        """
        Clip the values of the target array to fall inside the bounds (closed interval).

        Args:
            x: Numpy array to be clipped.

        Returns:
            Clipped numpy array with all its values inside the defined bounds.

        """
        return judo.clip(judo.astype(x, judo.float), self.low, self.high)

    def points_in_bounds(self, x: Tensor) -> Union[Tensor, bool]:
        """
        Check if the rows of the target array have all their coordinates inside \
        specified bounds.

        If the array is one dimensional it will return a boolean, otherwise a vector of booleans.

        Args:
            x: Array to be checked against the bounds.

        Returns:
            Numpy array of booleans indicating if a row lies inside the bounds.

        """
        match = self.clip(x) == judo.astype(x, judo.float)
        return match.all(1).flatten() if len(match.shape) > 1 else match.all()

    def safe_margin(
        self,
        low: Union[Tensor, Scalar] = None,
        high: Optional[Union[Tensor, Scalar]] = None,
        scale: float = 1.0,
    ) -> "Bounds":
        """
        Initialize a new :class:`Bounds` with its bounds increased o decreased \
        by an scale factor.

        This is done multiplying both high and low for a given factor. The value of the new \
        high and low will be proportional to the maximum and minimum values of \
        the array. Scale defines the proportion to make the bounds bigger and smaller. For \
        example, if scale is 1.1 the higher bound will be 10% higher, and the lower bounds 10% \
        smaller. If scale is 0.9 the higher bound will be 10% lower, and the lower bound 10% \
        higher. If scale is one, `high` and `low` will be equal to the maximum and minimum values \
        of the array.

        Args:
            high: Used to scale the `high` value of the current instance.
            low: Used to scale the `low` value of the current instance.
            scale: Value representing the tolerance in percentage from the current maximum and \
            minimum values of the array.

        Returns:
            :class:`Bounds` with scaled high and low values.

        """
        xmax = self.high if high is None else high
        xmin = self.low if low is None else low
        xmin_scaled, xmax_scaled = self.get_scaled_intervals(xmin, xmax, scale)
        return Bounds(low=xmin_scaled, high=xmax_scaled)

    def to_tuples(self) -> Tuple[Tuple[Scalar, Scalar], ...]:
        """
        Return a tuple of tuples containing the lower and higher bound for each \
        coordinate of the :class:`Bounds` shape.

        Returns:
            Tuple of the form ((x0_low, x0_high), (x1_low, x1_high), ...,\
              (xn_low, xn_high))
        Examples:
            >>> import numpy
            >>> array = numpy.array([1, 2, 5])
            >>> bounds = Bounds(high=array, low=-array)
            >>> print(bounds.to_tuples())
            ((-1, 1), (-2, 2), (-5, 5))

        """
        return tuple([dim for dim in zip(self.low, self.high)])
